from numba import jit
import numpy as np
from tifffile import imsave, imshow
from . import watershed, findmax
from scipy.ndimage import label, distance_transform_edt
from skimage.measure import regionprops

# from skimage.feature import peak_local_max
# from skimage.morphology import watershed
from tifffile.tifffile import TiffFile
from skimage.filters import threshold_otsu
import matplotlib.pyplot as plt
import os


@jit(nopython=True, parallel=True)
def find_focus(chunk):
    d, h, w = chunk.shape
    min_var = np.inf
    min_ind = int(d / 2)
    for i in range(d):
        tmp_v = np.std(chunk[i, :, :])
        if tmp_v < min_var:
            min_var = tmp_v
            min_ind = i
    return min_ind


@jit(nopython=True, parallel=True)
def get_chunk_mask(h, w, M, N):
    if h % M != 0 or w % N != 0:
        print("image can not be evenly divided")
    ch_h, ch_w = int(h / M), int(w / N)
    mask = np.array([1] * h * w).reshape(h, w)
    labels = list(range(M * N))
    count = 0
    for i in range(M):
        for j in range(N):
            mask[i * ch_h : (i + 1) * ch_h, j * ch_w : (j + 1) * ch_w] = labels[count]
            count += 1
    return mask


@jit(nopython=False, forceobj=True, parallel=True)
def get_focus_grid(img, mask):
    d, h, w = img.shape
    n_chunk = mask.max()
    focus_img = mask.copy()
    for ind in range(n_chunk + 1):
        min_var = np.inf
        min_ind = int(d / 2)
        cur_mask = mask == ind
        for i in range(d):
            # boolean slicing is not supported yet... so go python mode
            tmp_v = img[i, :, :][cur_mask].std()
            if tmp_v < min_var:
                min_var = tmp_v
                min_ind = i
        focus_img[cur_mask] = min_ind
    return focus_img


@jit(nopython=True, parallel=True)
def preCalculateParameters(
    first_ind=0, last_ind=33, N=101, direction=-1, zf=17, sigma=8.0
):
    h = (last_ind - first_ind) / N
    zs_d = np.array([first_ind + h * i for i in range(N)])
    zs_i = [int(i) for i in zs_d]
    return (
        h,
        zs_i,
        [
            direction
            * (zs_d[i] - zf)
            * np.exp(-(zf - zs_d[i]) ** 2 / (2 * (sigma ** 2)))
            for i in range(N)
        ],
    )


@jit(nopython=True)
def integrate(z_pile, h, zs_i, smooth_ponderation, first_ind=0, last_ind=33, N=101):
    total = (
        z_pile[zs_i[first_ind]] * smooth_ponderation[first_ind]
        + z_pile[zs_i[N - 1]] * smooth_ponderation[N - 1]
    )
    for i in range(1, N, 2):
        total += (
            4.0
            * z_pile[zs_i[first_ind + int(i)]]
            * smooth_ponderation[first_ind + int(i)]
        )
    for i in range(2, N - 1, 2):
        total += (
            2.0
            * z_pile[zs_i[first_ind + int(i)]]
            * smooth_ponderation[first_ind + int(i)]
        )
    return total * h / 3.0


@jit(nopython=True)
def compute(img, integrated, zf, focus_img, h, zs_i, smoothed_ponderation):
    depth, height, width = img.shape
    for x in range(height):
        for y in range(width):
            if focus_img[x, y] == zf:
                integrated[x, y] = integrate(
                    img[:, x, y], h, zs_i, smoothed_ponderation, last_ind=depth
                )


def segment(bf_stack, region_mask, zf_params):
    focus_img = get_focus_grid(bf_stack, region_mask)
    h, w = focus_img.shape
    for zf in np.unique(focus_img):
        if zf not in zf_params:
            zf_params[zf] = preCalculateParameters(zf=zf, last_ind=bf_stack.shape[0])
    integrated = np.zeros(h * w).reshape(h, w)
    for zf in np.unique(focus_img):
        compute(bf_stack, integrated, zf, focus_img, *zf_params[zf])
    th_otsu = threshold_otsu(integrated)
    return integrated, integrated <= th_otsu, zf_params


def get_master_fhs(root, pref):
    for r, ds, fs in os.walk(root):
        for f in fs:
            if (
                f.endswith(".tif")
                and f.startswith(pref)
                and not f.startswith(pref + "_end")
                and not f.endswith("_1.ome.tif")
            ):
                img_path = r + os.sep + f
                with TiffFile(img_path) as imgs:
                    if len(imgs.series) > 1:
                        yield r, imgs


def ipy_watershed(img, tol):
    dist = -distance_transform_edt(img)
    pts = findmax.find_maximum(dist, tol, False)
    buf = np.zeros(img.shape, dtype=np.uint16)
    buf[pts[:, 0], pts[:, 1]] = 1
    markers, n = label(buf, np.ones((3, 3)))
    line = watershed.watershed(dist, markers, line=True, conn=False + 1)
    img[line == 0] = 0
    return img


def skimage_watershed(img):
    # Generate the markers as local maxima of the distance to the background
    distance = distance_transform_edt(img)
    local_maxi = peak_local_max(
        distance, indices=False, footprint=np.ones((3, 3)), labels=img
    )
    markers = label(local_maxi)[0]
    return watershed(-distance, markers, mask=img, watershed_line=True)


def process_bf(
    r,
    img,
    region_mask,
    current_img_name,
    zf_params,
    show_img,
    seg_suffix="_SegAnalysis",
):
    res_dir = r + seg_suffix
    if not os.path.exists(res_dir):
        os.mkdir(res_dir)
    integrated, cur_mask, tmp_zf_params = segment(img, region_mask, zf_params)
    zf_params.update(tmp_zf_params)
    integrated = integrated.astype(np.int32)
    img_prefix = res_dir + os.sep + current_img_name
    imsave(img_prefix + "_integrated.tif", integrated)
    print("%s: integrated image saved!" % current_img_name)

    cur_mask = cur_mask.astype(np.uint8) * 255
    imsave(img_prefix + "_raw_mask.tif", cur_mask)
    print("%s: mask image saved!" % current_img_name)

    label_img, n = label(cur_mask, output=np.uint16)
    imsave(img_prefix + "_label.tif", label_img)
    print("%s: label image saved! %i rois detected" % (current_img_name, n))

    label_img, n = label(255 - cur_mask, output=np.uint16)
    cell_only_idx = (np.ones(n + 1) * 255).astype(np.uint8)
    pieces_only_idx = (np.ones(n + 1) * 255).astype(np.uint8)
    cell_only_idx[0] = 0
    pieces_only_idx[0] = 0
    res_features = regionprops(label_img, integrated)
    for p in res_features:
        if p.area < 500:
            cell_only_idx[p.label] = 0
        else:
            pieces_only_idx[p.label] = 0
    cells = cell_only_idx[label_img]
    imsave(img_prefix + "_clean.tif", cells)
    print("%s: cleaned mask image saved!" % current_img_name)
    imsave(img_prefix + "_rejected.tif", pieces_only_idx[label_img])
    print("%s: small pieces image saved!" % current_img_name)
    # watersheded_cells = skimage_watershed(255 - cells)
    watersheded_cells = ipy_watershed((255 - cells).copy(), 5)
    if show_img:
        imshow(watersheded_cells)
        plt.show()
    imsave(img_prefix + "_clean_watersheded.tif", watersheded_cells)
    print("%s: cleaned and watershed image saved!" % current_img_name)


def easy_run(root, M=4, N=4, h=2160, w=2560, show_img=False, zf_params={}):
    region_mask = get_chunk_mask(h, w, M, N)
    for r, imgs in get_master_fhs(root, "BF"):
        for img in imgs.series:
            process_bf(
                r,
                img.asarray(),
                region_mask,
                img[0].tags["MicroManagerMetadata"].value["FileName"][:-8],
                zf_params,
                show_img,
            )



import numpy as np
from numba import jit
from scipy.ndimage import label, generate_binary_structure
from skimage.io import imread, imsave


def neighbors(shape, conn=1):
    dim = len(shape)
    block = generate_binary_structure(dim, conn)
    block[tuple([1] * dim)] = 0
    idx = np.where(block > 0)
    idx = np.array(idx, dtype=np.uint8).T
    idx = np.array(idx - [1] * dim)
    acc = np.cumprod((1,) + shape[::-1][:-1])
    return np.dot(idx, acc[::-1])


@jit
def step(img, msk, line, pts, s, level, up, nbs):
    cur = 0
    while cur < s:
        p = pts[cur]
        if msk[p] == 0xFFFFFFFE or up and img[p] > level or not up and img[p] < level:
            cur += 1
            continue
        for dp in nbs:
            cp = p + dp
            if msk[cp] == 0:
                msk[cp] = msk[p]
                if s == len(pts):
                    s, cur = clear(pts, s, cur)
                pts[s] = cp
                s += 1

            elif msk[cp] == msk[p]:
                continue
            elif msk[cp] == 0xFFFFFFFF:
                continue
            elif msk[cp] == 0xFFFFFFFE:
                continue

            elif line:
                msk[cp] = 0xFFFFFFFE
        pts[cur] = -1
        cur += 1
    return cur


@jit
def clear(pts, s, cur):
    ns = 0
    nc = 0
    for c in range(s):
        if pts[c] != -1:
            pts[ns] = pts[c]
            ns += 1
            if c < cur:
                nc += 1
    return ns, nc


@jit
def collect(img, mark, nbs, pts):
    bins = np.zeros(img.max() + 1, dtype=np.uint32)
    cur = 0
    for p in range(len(mark)):
        bins[img[p]] += 1
        if mark[p] == 0xFFFFFFFF:
            continue  # edge
        if mark[p] == 0:
            continue  # zero
        for dp in nbs:
            if mark[p + dp] != mark[p]:
                pts[cur] = p
                cur += 1
                break
    return cur, bins


@jit
def erose(mark):
    for i in range(len(mark)):
        if mark[i] > 0xFFFFFFF0:
            mark[i] = 0


def buffer(img, dtype):
    buf = np.zeros(tuple(np.array(img.shape) + 2), dtype=dtype)
    buf[tuple([slice(1, -1)] * buf.ndim)] = img
    return buf


def watershed(img, mark, conn=1, line=False, up=True):
    maxv = img.max()
    minv = img.min()
    if img.dtype != np.uint8:
        img = ((img - minv) * (255 / (maxv - minv))).astype(np.uint8)
    img = buffer(img, np.uint8)
    mark = buffer(mark, np.uint32)
    ndim = img.ndim
    for n in range(ndim):
        idx = [slice(None) if i != n else [0, -1] for i in range(ndim)]
        mark[tuple(idx)] = 0xFFFFFFFF

    nbs = neighbors(img.shape, conn)
    img = img.ravel()
    mark1d = mark.ravel()

    pts = np.zeros(img.size // 3, dtype=np.int64)
    s, bins = collect(img, mark1d, nbs, pts)

    for level in range(len(bins))[:: 1 if up else -1]:
        if bins[level] == 0:
            continue
        s, c = clear(pts, s, 0)
        s = step(img, mark1d, line, pts, s, level, up, nbs)
    if line:
        erose(mark1d)
    return mark[tuple([slice(1, -1)] * ndim)]


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from skimage.filters import sobel
    from skimage.data import coins

    coins = coins()
    dem = sobel(coins)
    markers = np.zeros_like(coins, dtype=np.uint32)
    markers[coins < 30] = 1
    markers[coins > 150] = 2
    plt.imshow(markers)
    plt.show()
    watershed(dem, markers)
    plt.imshow(markers)
    plt.show()
    """
    from skimage.io import imread
    import matplotlib.pyplot as plt
    from time import time
    
    dem = imread('ice.png')
    mark = imread('mark.png')
    mark, n = label(mark>0, generate_binary_structure(2,2), output=np.uint16)
    start = time()
    #skrst = skwsh(dem, mark, watershed_line=True)
    print('skimage:', time()-start)
    watershed(dem, mark.copy())
    start = time()
    watershed(dem, mark, 1)
    print('mine:', time()-start)

    imsave('line.png', ((mark>0)*255).astype(np.uint8))
    """



import scipy.ndimage as ndimg
import numpy as np
from numba import jit


def neighbors(shape):
    dim = len(shape)
    block = np.ones([3] * dim)
    block[tuple([1] * dim)] = 0
    idx = np.where(block > 0)
    idx = np.array(idx, dtype=np.uint8).T
    idx = np.array(idx - [1] * dim)
    acc = np.cumprod((1,) + shape[::-1][:-1])
    return np.dot(idx, acc[::-1])


@jit  # trans index to r, c...
def idx2rc(idx, acc):
    rst = np.zeros((len(idx), len(acc)), dtype=np.int16)
    for i in range(len(idx)):
        for j in range(len(acc)):
            rst[i, j] = idx[i] // acc[j]
            idx[i] -= rst[i, j] * acc[j]
    return rst


@jit  # fill a node (may be two or more points)
def fill(img, msk, p, nbs, buf):
    buf[0] = p
    back = img[p]
    cur = 0
    s = 1

    while cur < s:
        p = buf[cur]
        for dp in nbs:
            cp = p + dp
            if img[cp] == back and msk[cp] == 1:
                msk[cp] = 3
                buf[s] = cp
                s += 1
                if s == len(buf):
                    buf[: s - cur] = buf[cur:]
                    s -= cur
                    cur = 0
        cur += 1


@jit(forceobj=True)  # my mark
def mark(img, msk, buf, mode):  # mark the array use (0, 1, 2)
    nbs = neighbors(img.shape)
    idx = np.zeros(msk.size // 3, dtype=np.int64)
    img = img.ravel()
    msk = msk.ravel()
    s = 0
    for p in range(len(img)):
        if msk[p] != 1:
            continue
        sta = 0
        for dp in nbs:
            if img[p + dp] == img[p]:
                sta += 1
            if mode and img[p + dp] > img[p]:
                sta = 100
                break
            elif not mode and img[p + dp] < img[p]:
                sta = 100
                break
        if sta == 100:
            continue
        msk[p] = 3
        if sta > 0:
            fill(img, msk, p, nbs, buf)

        idx[s] = p
        s += 1
        if s == len(idx):
            break
    return idx[:s].copy()


@jit(nopython=False, forceobj=True)  # 3 max 2 zmd b4 ptd
def filter(img, msk, idx, bur, tor, mode):
    nbs = neighbors(img.shape)
    acc = np.cumprod((1,) + img.shape[::-1][:-1])[::-1]
    img = img.ravel()
    msk = msk.ravel()

    arg = np.argsort(img[idx])[:: -1 if mode else 1]

    for i in arg:
        if msk[idx[i]] != 3:
            idx[i] = 0
            continue
        cur = 0
        s = 1
        bur[0] = idx[i]
        while cur < s:
            p = bur[cur]
            if msk[p] == 2:
                idx[i] = 0
                break

            for dp in nbs:
                cp = p + dp
                if msk[cp] == 0 or cp == idx[i] or msk[cp] == 4:
                    continue
                if mode and img[cp] < img[idx[i]] - tor:
                    continue
                if not mode and img[cp] > img[idx[i]] + tor:
                    continue
                bur[s] = cp
                s += 1
                if s == msk.size // 3:
                    cut = cur // 2
                    msk[bur[:cut]] = 2
                    bur[: s - cut] = bur[cut:]
                    cur -= cut
                    s -= cut

                if msk[cp] != 2:
                    msk[cp] = 4
            cur += 1
        msk[bur[:s]] = 2

    return idx2rc(idx[idx > 0], acc)


def find_maximum(img, tor, mode=True):
    msk = np.zeros_like(img, dtype=np.uint8)
    msk[tuple([slice(1, -1)] * img.ndim)] = 1
    buf = np.zeros(img.size // 3, dtype=np.int64)
    idx = mark(img, msk, buf, mode)
    idx = filter(img, msk, idx, buf, tor, mode)
    return idx


if __name__ == "__main__":
    from skimage.io import imread
    from scipy.ndimage import gaussian_filter
    from time import time
    import matplotlib.pyplot as plt

    img = gaussian_filter(imread("test.png"), 0)
    pts = find_maximum(img, 20, True)
    start = time()
    pts = find_maximum(img, 10, True)
    print(time() - start)
    plt.imshow(img, cmap="gray")
    plt.plot(pts[:, 1], pts[:, 0], "y.")
    plt.show()



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="bfseg",
    version="0.0.6",
    description="a segmentation algorithm to find cell borders in bright-field images",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Tong LI",
    author_email="tongli.bioinfo@gmail.com",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        #'Topic :: segmentation :: Microscope',
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords="microscope segmentation bright-field binary",
    packages=find_packages(),
    python_requires=">=3.0.*",
    # install_requires=['numba', 'numpy', 'skimage'],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    # entry_points={  # Optional
    #    'console_scripts': [
    #        'sample=sample:main',
    #    ],
    # },
    # project_urls={  # Optional
    #    'Bug Reports': 'https://github.com/pypa/sampleproject/issues',
    #    'Funding': 'https://donate.pypi.org',
    #    'Say Thanks!': 'http://saythanks.io/to/example',
    #    'Source': 'https://github.com/pypa/sampleproject/',
    # }
)
